# Copyright (c) Sandro Tosi, 2023, under the terms of dput-ng.

# Installation & setup guide (similar to the twitter hook, refer to that if unclear)

# 1. register an application on the Mastodon instance you want to publish messages to
# 2. save the application access token and instance API endpoint in ~/.config/dput-ng/mastodon.json
#    (sample at skel/mastodon.json)
# 3. run `sudo apt install python3-mastodon`
# 4. set up the hook by copying skel/toot.json to ~/.dput.d/hooks/
#    and code/toot.py to ~/.dput.d/scripts/
# 5. edit ~/.dput.d/profiles/ftp-master.json to add "toot" as an extra hook

import json
import os

from mastodon import Mastodon


def toot(changes, profile, interface):
    toot = "I've just uploaded %s/%s to %s #debian" % (
        changes['Source'],
        changes['Version'],
        changes['Distribution']
    )
    # https://fedi.tips/why-do-some-people-on-mastodon-and-the-fediverse-have-bigger-character-limits/
    if len(toot) > 500:
        toot = toot[:500]

    with open(os.path.expanduser("~/.config/dput-ng/mastodon.json"), 'r') as f:
        obj = json.load(f)
    m = Mastodon(
        access_token=obj["access_token"],
        api_base_url=obj["api_base_url"]
    )

    m.toot(toot)



from dput.util import run_command
from dput import upload
from dput.exceptions import UploadException
import dput.core
import glob
import os.path
import os
import pytest

dput.core.CONFIG_LOCATIONS = {
    os.path.abspath("./tests/dputng"): 0
}  # Isolate.


@pytest.fixture
def monkeypatch_ubuntu_supported(monkeypatch):
    """Patch the set of supported Ubuntu codenames."""
    monkeypatch.setattr(
        "distro_info.UbuntuDistroInfo.supported",
        lambda *args, **kwargs: ["xenial", "bionic", "focal", "jammy"],
    )


@pytest.fixture
def isolate_uploads(monkeypatch, tmp_path):
    """
    Isolate uploads to a temporary directory.

    We don't want uploads to go anywhere real.  Our test profiles all use
    the "local" method, but this also arranges for the uploads to end up in
    a temporary directory whose contents can then be inspected by tests.
    """
    # Take advantage of the fact that LocalUploader.upload_file calls
    # os.path.expanduser.
    def fake_expanduser(arg):
        if arg == "/dev/null":
            return str(tmp_path)
        else:
            return str(tmp_path / arg)

    monkeypatch.setattr(os.path, "expanduser", fake_expanduser)
    yield tmp_path


def _build_fnord(version='1.0'):
    popdir = os.path.abspath(os.getcwd())
    os.chdir("tests/fake_package/fake-package-%s" % version)
    cmd = "dpkg-buildpackage -us -uc -S"
    env = {
        "DEB_VENDOR": "Ubuntu",
        "DPKG_ORIGINS_DIR": "../../dpkg-origins",
        "PATH": os.environ["PATH"],
    }
    stdout, stderr, ret = run_command(cmd, env=env)
    upload_files = glob.glob("../fnord_%s_source.*.upload" % version)
    for fn in upload_files:
        os.unlink(fn)
    os.chdir(popdir)
    if ret != 0:
        print('Package build failed.')
        print(f'cmd: {cmd}')
        print(f'env: {env}')
        print('###### stdout:')
        print(stdout)
        print('###### stderr:')
        print(stderr)
        print('###### - end log')
        assert False
    return os.path.abspath("tests/fake_package/fnord_%s_source.changes"
                           % version)


def test_upload(monkeypatch_ubuntu_supported, isolate_uploads):
    """ Test the upload of a package """
    path = _build_fnord()
    upload(path, 'test')
    assert (isolate_uploads / "fnord_1.0_source.changes").exists()


def test_double_upload(monkeypatch_ubuntu_supported, isolate_uploads):
    """ Test a double-upload (and force block) """
    path = _build_fnord()
    upload(path, 'test')
    assert (isolate_uploads / "fnord_1.0_source.changes").exists()
    with pytest.raises(
        UploadException, match="Package fnord was already uploaded to test"
    ):
        upload(path, 'test')


def test_ppa_upload(monkeypatch_ubuntu_supported, isolate_uploads):
    """ Test the upload of a package to a PPA (no Launchpad-Bugs-Fixed) """
    path = _build_fnord(version='1.1')
    (isolate_uploads / "~foo/bar").mkdir(parents=True)
    upload(path, 'ppa:foo/bar')
    assert (isolate_uploads / "~foo/bar/fnord_1.1_source.changes").exists()



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4

# Copyright (c) 2012, 2024 dput authors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your Option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
"""
Common functions for SSH-based uploaders
"""

import os
import paramiko
import pwd

from dput.exceptions import UploadException


def get_ssh_config(fqdn):
    config = paramiko.SSHConfig()
    if os.path.exists('/etc/ssh/ssh_config'):
        with open('/etc/ssh/ssh_config') as f:
            config.parse(f)
    if os.path.exists(os.path.expanduser('~/.ssh/config')):
        with open(os.path.expanduser('~/.ssh/config')) as f:
            config.parse(f)
    return config.lookup(fqdn)


def find_username(conf, ssh_conf):
    """
    Given a profile (``conf``) and an SSH configuration (``ssh_conf``),
    return the preferred username to login with.
    The profile takes precedence over the SSH configuration.
    It falls back to getting the logged in user's name.
    """
    user = None
    user = pwd.getpwuid(os.getuid()).pw_name
    if 'user' in ssh_conf:
        user = ssh_conf['user']
    if 'login' in conf:
        new_user = conf['login']
        if new_user != '*':
            user = new_user
    if not user:
        raise UploadException(
            "No user to upload could be retrieved. "
            "Please set 'login' explicitly in your profile"
        )
    return user



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4

# Copyright (c) 2012, 2023 dput authors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your Option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
"""
Rsync Uploader implementation
"""

import os.path

from dput.core import logger
from dput.uploader import AbstractUploader
from dput.exceptions import UploadException
from dput.uploaders._ssh_common import get_ssh_config, find_username
from dput.util import run_command


class RsyncUploadException(UploadException):
    """
    Thrown in the event of a problem connecting, uploading to or
    terminating the connection with the remote server. This is
    a subclass of :class:`dput.exceptions.UploadException`.
    """
    pass


class RsyncUploader(AbstractUploader):
    """
    Provides an interface to upload files through rsync.

    This is a subclass of :class:`dput.uploader.AbstractUploader`
    """

    def initialize(self, **kwargs):
        """
        See :meth:`dput.uploader.AbstractUploader.initialize`
        """
        ssh_base = ["ssh", "-x"]
        if "port" in self._config:
            ssh_base += ["-p", self._config["port"]]
        self._rsync_base = ["rsync", "--copy-links", "--partial", "-za",
                            "--chmod=644", "-e", " ".join(ssh_base)]
        fqdn = self._config["fqdn"]
        self._rsync_host = "%s@%s" % (find_username(self._config,
                                                    get_ssh_config(fqdn)),
                                      fqdn)
        logger.debug("Using rsync to upload to %s" % (self._rsync_host))

    def upload_file(self, filename, upload_filename=None):
        """
        See :meth:`dput.uploader.AbstractUploader.upload_file`
        """

        if not upload_filename:
            upload_filename = os.path.basename(filename)

        incoming = self._config["incoming"]
        targetfile = "%s:%s" % (self._rsync_host,
                                os.path.join(incoming, upload_filename))
        rsync = self._rsync_base + [filename, targetfile]
        (_, e, x) = run_command(rsync)
        if x != 0:
            raise RsyncUploadException("Failed to upload %s to %s: %s" % (
                upload_filename, targetfile, e)
            )

    def shutdown(self):
        """
        See :meth:`dput.uploader.AbstractUploader.shutdown`
        """
        pass



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: GPL-2+
# Copyright 2023 Helmut Grohne

import re

import dput.command
import dput.commands.dm
import dput.exceptions


class MigrateException(dput.exceptions.DcutError):
    pass


class MigrateCommand(dput.command.AbstractCommand):
    def __init__(self, interface):
        super().__init__(interface)
        self.cmd_name = "migrate"
        self.cmd_purpose = "migrate or block a staged upload"

    def name_and_purpose(self):
        return (self.cmd_name, self.cmd_purpose)

    def register(self, parser, **kwargs):
        parser.add_argument(
            "--reject",
            default=False,
            action="store_true",
            help="reject the upload rather than accepting it",
        )
        parser.add_argument(
            "source",
            action="store",
            help="source package to operate on",
        )
        parser.add_argument(
            "version",
            action="store",
            help="source version to operate on",
        )

    def validate(self, args):
        if not re.match("^[a-z0-9][a-z0-9.+-]+$", args.source):
            raise MigrateException("invalid source package name")
        if not re.match("^[A-Za-z0-9.+:~-]+$", args.version):
            raise MigrateException("invalid source version")

    def generate_commands_name(self, profile):
        return dput.commands.dm.generate_dak_commands_name(profile)

    def produce(self, fh, args):
        command = "REJECT" if args.reject else "ACCEPT"
        # The initial empty line is important.
        fh.write(f"""
Action: process-upload
Command: {command}
Source: {args.source}
Version: {args.version}
""")



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4

# Copyright (c) 2012 dput authors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your Option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
"""
SFTP Uploader implementation
"""

import paramiko
import socket
import errno
import os.path
from binascii import hexlify

from dput.core import logger
from dput.uploader import AbstractUploader
from dput.uploaders._ssh_common import get_ssh_config, find_username
from dput.exceptions import UploadException


class SftpUploadException(UploadException):
    """
    Thrown in the event of a problem connecting, uploading to or
    terminating the connection with the remote server. This is
    a subclass of :class:`dput.exceptions.UploadException`.
    """
    pass


def check_paramiko_version(req):
    """
    Return whether paramiko satisfies the given a version requirement (``req``),
    """
    try:
        # Prefer __version_info__ over parsing __version__ on our own although
        # it is missing in some release of paramiko.
        version_info = paramiko.__version_info__
    except AttributeError:
        version_info = tuple(int(value) for value in
                             paramiko.__version__.split('.'))
    return version_info >= req


class AskToAccept(paramiko.AutoAddPolicy):
    """
    Paramiko policy to automatically add the hostname, but only after asking.
    """

    def __init__(self, uploader):
        super(AskToAccept, self).__init__()
        self.uploader = uploader

    def missing_host_key(self, client, hostname, key):
        accept = self.uploader.interface.boolean(
            title='please login',
            message='To accept %s hostkey %s for %s type "yes":' % (
                key.get_name(),
                hexlify(key.get_fingerprint()),
                hostname
            )
        )
        if accept:
            super(AskToAccept, self).missing_host_key(client, hostname, key)
        else:
            raise paramiko.SSHException('Unknown server %s' % hostname)


class SFTPUploader(AbstractUploader):
    """
    Provides an interface to upload files through SFTP.

    This is a subclass of :class:`dput.uploader.AbstractUploader`
    """

    def initialize(self, **kwargs):
        """
        See :meth:`dput.uploader.AbstractUploader.initialize`
        """
        fqdn = self._config['fqdn']
        incoming = self._config['incoming']

        self.sftp_config = {}
        if "sftp" in self._config:
            self.sftp_config = self._config['sftp']

        self.putargs = {'confirm': False}

        if "confirm_upload" in self.sftp_config:
            self.putargs['confirm'] = self.sftp_config['confirm_upload']

        if incoming.startswith('~/'):
            logger.warning("SFTP does not support ~/path, continuing with"
                           "relative directory name instead.")
            incoming = incoming[2:]
        # elif incoming.startswith('~') and not self.host_is_launchpad:
        #    raise SftpUploadException("SFTP doesn't support ~path. "
        #                              "if you need $HOME paths, use SCP.")
        #  XXX: What to do here?? - PRT

        ssh_kwargs = {
            "port": 22,
            "compress": True
        }

        # XXX: Timeout override

        if 'port' in self._config:
            ssh_kwargs['port'] = self._config['port']

        if 'scp_compress' in self._config:
            ssh_kwargs['compress'] = self._config['scp_compress']

        o = get_ssh_config(fqdn)

        user = find_username(self._config, o)

        ssh_kwargs['username'] = user

        if 'identityfile' in o:
            if check_paramiko_version((1, 10)):
                # Starting with paramiko 1.10 identityfile is always a list.
                pkey = [os.path.expanduser(path) for path in o['identityfile']]
            else:
                pkey = os.path.expanduser(o['identityfile'])
            ssh_kwargs['key_filename'] = pkey

        logger.info("Logging into host %s as %s" % (fqdn, user))
        self._sshclient = paramiko.SSHClient()
        if 'globalknownhostsfile' in o:
            for gkhf in o['globalknownhostsfile'].split():
                if os.path.isfile(gkhf):
                    self._sshclient.load_system_host_keys(gkhf)
        else:
            files = [
                "/etc/ssh/ssh_known_hosts",
                "/etc/ssh/ssh_known_hosts2"
            ]
            for fpath in files:
                if os.path.isfile(fpath):
                    self._sshclient.load_system_host_keys(fpath)

        if 'userknownhostsfile' in o:
            for u in o['userknownhostsfile'].split():
                # actually, ssh supports a bit more than ~/,
                # but that would be a task for paramiko...
                ukhf = os.path.expanduser(u)
                if os.path.isfile(ukhf):
                    self._sshclient.load_host_keys(ukhf)
        else:
            for u in ['~/.ssh/known_hosts2', '~/.ssh/known_hosts']:
                ukhf = os.path.expanduser(u)
                if os.path.isfile(ukhf):
                    # Ideally, that should be load_host_keys,
                    # so that the known_hosts file can be written
                    # again. But paramiko can destroy the contents
                    # or parts of it, so no writing by using
                    # load_system_host_keys here, too:
                    self._sshclient.load_system_host_keys(ukhf)
        self._sshclient.set_missing_host_key_policy(AskToAccept(self))
        self._auth(fqdn, ssh_kwargs)
        try:
            self._sftp = self._sshclient.open_sftp()
        except paramiko.SSHException as e:
            raise SftpUploadException(
                "Error opening SFTP channel to %s (perhaps sftp is "
                "disabled there?): %s" % (
                    fqdn,
                    repr(e)
                )
            )

        # logger.debug("Changing directory to %s" % (incoming))
        # self._sftp.chdir(incoming)
        try:
            self._sftp.stat(incoming)
        except IOError as e:
            # launchpad does not support any operations to check if a directory
            # exists. stat will fail with an IOError with errno equal to None.
            if e.errno is None:
                logger.debug(
                    "Failed to stat incoming directory %s on %s. This should "
                    "only happen on launchpad." % (
                        incoming,
                        fqdn
                    )
                )
            else:
                raise SftpUploadException(
                    "Failed to stat incoming directory %s on %s: %s" % (
                        incoming,
                        fqdn,
                        e.strerror
                    )
                )
        except paramiko.SSHException as e:
            raise SftpUploadException("SFTP error uploading to %s: %s" % (
                fqdn,
                repr(e)
            ))
        self.incoming = incoming

    def _auth(self, fqdn, ssh_kwargs, _first=0):
        if _first == 3:
            raise SftpUploadException("Failed to authenticate")
        try:
            self._sshclient.connect(fqdn, **ssh_kwargs)
            logger.debug("Logged in!")
        except socket.error as e:
            raise SftpUploadException("SFTP error uploading to %s: %s" % (
                fqdn,
                repr(e)
            ))
        except paramiko.AuthenticationException:
            logger.warning("Failed to auth. Prompting for a login pair.")
            # XXX: Ask for pw only
            user = self.interface.question('please login', 'Username')
            # 4 first error
            pw = self.interface.password(None, "Password")

            if user is not None:
                ssh_kwargs['username'] = user
            ssh_kwargs['password'] = pw
            self._auth(fqdn, ssh_kwargs, _first=_first + 1)
        except paramiko.SSHException as e:
            raise SftpUploadException("SFTP error uploading to %s: %s" % (
                fqdn,
                repr(e)
            ))

    def upload_file(self, filename, upload_filename=None):
        """
        See :meth:`dput.uploader.AbstractUploader.upload_file`
        """

        if not upload_filename:
            upload_filename = os.path.basename(filename)

        upload_filename = os.path.join(self.incoming, upload_filename)
        logger.debug("Writing to: %s" % (upload_filename))

        try:
            self._sftp.put(filename, upload_filename, **self.putargs)
        except IOError as e:
            if e.errno == errno.EACCES:
                self.upload_write_error(e)
            else:
                raise SftpUploadException("Could not upload file %s: %s" % (
                    filename,
                    e
                ))

    def shutdown(self):
        """
        See :meth:`dput.uploader.AbstractUploader.shutdown`
        """
        self._sshclient.close()
        self._sftp.close()



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4

# Copyright (c) 2012 dput authors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
"""
Local Uploader implementation
"""

from dput.uploader import AbstractUploader
from dput.exceptions import UploadException
import dput.util
import os.path


class LocalUploadException(UploadException):
    """
    Thrown in the event of a problem installing to a local file.
    This is a subclass of :class:`dput.exceptions.UploadException`.
    """
    pass


class LocalUploader(AbstractUploader):
    """
    Provides an interface to "upload" files to the local filesystem. This
    is helpful when you're dputting to the same system you're currently on,
    and do not wish to use `scp` or `sftp` as the transport (which is totally
    understandable).
    """

    def initialize(self, **kwargs):
        """
        See :meth:`dput.uploader.AbstractUploader.initialize`
        """
        pass

    def upload_file(self, filename, upload_filename=None):
        """
        See :meth:`dput.uploader.AbstractUploader.upload_file`
        """

        whereto = self._config['incoming']
        whereto = os.path.expanduser(whereto)
        if "HOME" in os.environ:
            whereto = os.path.join(os.environ["HOME"], whereto)
        if upload_filename is not None:
            whereto = os.path.join(whereto, upload_filename)
        command = [
            "install",
            "--mode",
            "0644",
            filename,
            whereto
        ]
        (_, e, x) = dput.util.run_command(command)
        if x != 0:
            raise LocalUploadException("Command %s returned failure %d: %s" % (
                repr(command), x, e)
            )

    def shutdown(self):
        """
        See :meth:`dput.uploader.AbstractUploader.shutdown`
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4

# Copyright (c) 2012 dput authors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from dput.profile import apply_overrides, parse_overrides
from dput.exceptions import DputConfigurationError

def test_parse_overrides():
    test_cases = [
        ({'foo': ['bar']},
            ['foo=bar']),
        ({'foo': [None]},
            ['-foo']),
        ({'+foo': ['bar']},
            ['+foo=bar']),
        ({'foo': [None, '1']},
            ['-foo', 'foo=1']),
        ({'foo': {'bar': {'baz': ['paultag.is.a.god']}}},
            ['foo.bar.baz=paultag.is.a.god']),
        ({'foo': {'bar': ['1', '2', '3']}, 'foo2': ['4']},
            ['foo.bar=1', 'foo.bar=2', 'foo.bar=3', 'foo2=4']),
        ({'foo': {'bar': ['"True=True"', "'foo'=False 'obj2'"]},
          'foo2': ['bar']},
            ['foo.bar="True=True"', "foo.bar='foo'=False 'obj2'",
             'foo2=bar'])
     ]

    bad_test_cases = [
                      (None, ['invalid']),
                      (None, ['foo=bar', 'foo=bar', 'foo.bar=bar']),
                      (None, ['foo.bar=1', 'foo.bar.baz=1'])
    ]

    for (expectation, case) in test_cases:
        print(parse_overrides(case))
        assert(expectation == parse_overrides(case))

    for (expectation, case) in bad_test_cases:
        try:
            parse_overrides(case)
            raise Exception("Bad test case: %s" % (case))
        except DputConfigurationError:
            pass


def test_apply_overrides():
    test_cases = [
        ({'hooks': ['foo', 'bar']},
         {'hooks': ['foo']},
         {'hooks': ['bar']}),
        ({'hooks': ['bar']},
         {'hooks': ['foo']},
         {'hooks': [None, 'bar']}),
        ({'hooks': ['foo'], 'plain': 'bar'},
         {'hooks': ['foo'], 'plain': 'foo'},
         {'plain': ['bar']}),
        ({'new': 'foo'},
         {},
         {'new': ['foo']}),
        ({},
         {'previous': 'foo'},
         {'previous': [None]}),
    ]

    for expectation, profile, overrides in test_cases:
        profile = profile.copy()
        apply_overrides(profile, overrides)
        assert expectation == profile
